/*
	Phoenix Suite 2
	Author: Jean-Baptiste Wautier
	Copyright 2006-2009 (c)
*/
namespace Phoenix.Log.Reader
{
    using System;
    using System.Collections.Generic;
    using System.Data;
    using System.Text;

    using Phoenix.Log;
    using Phoenix.Log.Reader;
    using Phoenix.Log.Writer;

    /// <summary>
    /// Represents features to read the log of an IAppender
    /// </summary>
    public class SQLiteLogReader : ILogReader
    {
        #region Fields

        private uint count;
        private string databasePath = string.Empty;
        private DateTime? maxDate = null;
        private DateTime? minDate = null;

        #endregion Fields

        #region Constructors

        /// <summary>
        /// Initializes assembly new instance of the <see cref="SQLiteLogReader"/> class.
        /// </summary>
        /// <param name="databasePath">The connection string.</param>
        internal SQLiteLogReader(string databasePath)
        {
            this.databasePath = databasePath;
        }

        #endregion Constructors

        #region Properties

        /// <summary>
        /// Gets the log count.
        /// </summary>        
        public uint LogCount
        {
            get
            {
                return this.count;
            }
        }

        /// <summary>
        /// Gets the date of the newest log.
        /// </summary>        
        public DateTime NewestLog
        {
            get
            {
                return this.minDate.Value;
            }
        }

        /// <summary>
        /// Gets the date of the oldest log.
        /// </summary>        
        public DateTime OldestLog
        {
            get
            {
                return this.maxDate.Value;
            }
        }

        #endregion Properties

        #region Methods

        /// <summary>
        /// Gets all the logs.
        /// </summary>
        /// <returns>The list of logs</returns>
        public List<IRow> GetAllLogs()
        {
            string storedProcedure = StoredProcedures.SELECT_LogSorted;
            List<IRow> list = new List<IRow>();

            SQLitePersistor persistor = new SQLitePersistor(this.databasePath);
            persistor.LoadStoredProcedure(storedProcedure);

            using (DataSet dataSet = persistor.ExecuteQuery())
            {

                if (dataSet.Tables.Count == 0)
                    return null;

                foreach (DataRow dataRow in dataSet.Tables[0].Rows)
                {
                    list.Add(GetLogRowFromDataRow(dataRow));
                }
            }
            return list;
        }

        /// <summary>
        /// Gets the dependency of the log with the specified Guid.
        /// </summary>
        /// <param name="parent">The GUID.</param>
        /// <returns>
        /// The list of the dependency of the specified log
        /// </returns>
        public List<IRow> GetChildren(Guid parent)
        {
            List<IRow> list = new List<IRow>();
            IPersistor persistor = new SQLitePersistor(this.databasePath);
            string storedProcedure = StoredProcedures.SELECT_Dependency;

            persistor.LoadStoredProcedure(storedProcedure);
            persistor.Arguments.Add("ID", "'" + parent.ToString("D") + "'");

            using (DataSet dataSet = persistor.ExecuteQuery())
            {

                if (dataSet.Tables.Count == 0)
                    return list;

                foreach (DataRow dataRow in dataSet.Tables[0].Rows)
                {
                    IRow item = GetLogRowFromDataRow(dataRow);
                    item.BaseLog = GetLogById(item.BaseLogKey);
                    list.Add(item);
                }
            }
            return list;
        }

        /// <summary>
        /// Gets the dependency of the log with the specified Guid.
        /// </summary>
        /// <param name="parent">The parent.</param>
        /// <returns>
        /// The list of the dependency of the specified log
        /// </returns>
        /// <exception cref="OverflowException">The format of parent is invalid.</exception>
        /// <exception cref="FormatException">The format of parent is invalid.</exception>
        public List<IRow> GetChildren(string parent)
        {
            Guid realGuid;

            if (string.IsNullOrEmpty(parent))
                throw new ArgumentNullException("parent", "The specified guid is null or empty!");

            realGuid = new Guid(parent);
            return this.GetChildren(realGuid);
        }

        /// <summary>
        /// Gets the dependency of the specified log.
        /// </summary>
        /// <param name="baseLog">The base log.</param>
        /// <returns>
        /// The list of the dependency of the specified log
        /// </returns>
        public List<IRow> GetChildren(IRow baseLog)
        {
            if (baseLog == null)
                throw new ArgumentNullException("baseLog", "Impossible to get the children of a null parent!");
            return GetChildren(baseLog.Key);
        }

        /// <summary>
        /// Gets the row with the specified Guid.
        /// </summary>
        /// <param name="id">The GUID.</param>
        /// <returns>The row with the specified Guid</returns>
        public IRow GetLogById(Guid id)
        {
            return this.GetLogById(id.ToString("D"));
        }

        /// <summary>
        /// Gets the row.
        /// </summary>
        /// <param name="id">The GUID.</param>
        /// <returns>The row with the specified Guid</returns>
        public IRow GetLogById(string id)
        {
            IPersistor persistor = new SQLitePersistor(this.databasePath);
            IRow logRow;

            persistor.LoadStoredProcedure(StoredProcedures.SELECT_LogOnID);
            persistor.Arguments.Add("ID", "'" + id + "'");
            using (DataSet dataSet = persistor.ExecuteQuery())
            {

                if (dataSet.Tables.Count == 0 || dataSet.Tables[0].Rows.Count == 0)
                    return null;

                logRow = GetLogRowFromDataRow(dataSet.Tables[0].Rows[0]);
            }
            return logRow;
        }

        /// <summary>
        /// Gets the logs between.
        /// </summary>
        /// <param name="lowerThreshold">The lower threshold.</param>
        /// <param name="upperThreshold">The upper threshold.</param>
        /// <returns>Search result.</returns>
        public List<IRow> GetLogsBetween(DateTime lowerThreshold, DateTime upperThreshold)
        {
            string fLowerThreshold = lowerThreshold.ToString("yyyy'-'MM'-'dd' 'HH:mm:ss");
            string fUpperThreshold = upperThreshold.ToString("yyyy'-'MM'-'dd' 'HH:mm:ss");
            IPersistor persistor = new SQLitePersistor(this.databasePath);
            persistor.LoadStoredProcedure(StoredProcedures.SELECT_LogsBetweenDate);
            persistor.Arguments.Add("LowerThreshold", fLowerThreshold);
            persistor.Arguments.Add("UpperThreshold", fUpperThreshold);
            return GetListFromDataSet(persistor.ExecuteQuery());
        }

        /// <summary>
        /// Gets the logs between.
        /// </summary>
        /// <param name="lowerThreshold">The lower threshold. Should have that format: yyyy-MM-dd HH:mm:ss</param>
        /// <param name="upperThreshold">The upper threshold. Should have that format: yyyy-MM-dd HH:mm:ss</param>
        /// <returns>Search result.</returns>
        public List<IRow> GetLogsBetween(string lowerThreshold, string upperThreshold)
        {
            string msg = "Impossible to cast the {0} threshold ['{1}'] into a DateTime!";
            DateTime dtLowerThreshold;
            DateTime dtUpperThreshold;
            if (!DateTime.TryParse(lowerThreshold, out dtLowerThreshold))
            {
                msg = string.Format(msg, "lower", lowerThreshold);
                throw new InvalidCastException(msg);
            }
            if (!DateTime.TryParse(upperThreshold, out dtUpperThreshold))
            {
                msg = string.Format(msg, "upper", upperThreshold);
                throw new InvalidCastException(msg);
            }
            return this.GetLogsBetween(dtLowerThreshold, dtUpperThreshold);
        }

        /// <summary>
        /// Gets all the logs of the specified type.
        /// </summary>
        /// <param name="stamp"></param>
        /// <returns>Search result</returns>
        public List<IRow> GetLogsBy(Phoenix.Log.Stamp stamp)
        {
            Log.Stamp[] logTypes = new Log.Stamp[] { stamp };
            return GetLogsBy(logTypes);
        }

        /// <summary>
        /// Gets all the logs of the specified list of types.
        /// </summary>
        /// <param name="logTypes">Stamp of the log.</param>
        /// <returns>Search result</returns>
        public List<IRow> GetLogsBy(Phoenix.Log.Stamp[] logTypes)
        {
            string arguments = string.Empty;
            foreach (Log.Stamp item in logTypes)
            {
                arguments += ((int)item).ToString() + ",";
            }
            arguments = arguments.TrimEnd(',');

            IPersistor persistor = new SQLitePersistor(this.databasePath);
            persistor.LoadStoredProcedure(StoredProcedures.SELECT_LogsByType);
            persistor.Arguments.Add("Type", arguments);

            return GetListFromDataSet(persistor.ExecuteQuery());
        }

        /// <summary>
        /// Gets the logs that contains the specified text into the exception message.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns></returns>
        public List<IRow> GetLogsWithExceptionMessage(string value)
        {
            IPersistor persistor = new SQLitePersistor(this.databasePath);
            persistor.LoadStoredProcedure(StoredProcedures.SELECT_LogsOnExceptionMessageContains);
            persistor.Arguments.Add("ExceptionMessage", value);

            return GetListFromDataSet(persistor.ExecuteQuery());
        }

        /// <summary>
        /// Gets the logs that contains the specified text into the message.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns></returns>
        public List<IRow> GetLogsWithMessage(string value)
        {
            IPersistor persistor = new SQLitePersistor(this.databasePath);
            persistor.LoadStoredProcedure(StoredProcedures.SELECT_LogsOnMessageContains);
            persistor.Arguments.Add("Message", value);

            return GetListFromDataSet(persistor.ExecuteQuery());
        }

        /// <summary>
        /// Gets the logs that contains the specified text into the title.
        /// </summary>
        /// <param name="value">The value.</param>
        /// <returns>Search result</returns>
        public List<IRow> GetLogsWithTitle(string value)
        {
            IPersistor persistor = new SQLitePersistor(this.databasePath);
            persistor.LoadStoredProcedure(StoredProcedures.SELECT_LogsOnTitleContains);
            persistor.Arguments.Add("Title", value);

            return GetListFromDataSet(persistor.ExecuteQuery());
        }

        /// <summary>
        /// Gets the titles.
        /// </summary>
        /// <returns>
        /// A list of the different titles in the logs
        /// </returns>
        public string[] GetTitles()
        {
            IPersistor persistor = new SQLitePersistor(this.databasePath);
            persistor.LoadStoredProcedure(StoredProcedures.SELECT_DistinctTitle);

            List<string> list = new List<string>();

            using (DataSet dataSet = persistor.ExecuteQuery())
            {

                if (dataSet.Tables.Count <= 0)
                    return null;

                foreach (DataRow item in dataSet.Tables[0].Rows)
                {
                    list.Add((string)item["Title"]);
                }
            }
            if (list.Count <= 0)
                return null;
            else
                return list.ToArray();
        }

        /// <summary>
        /// Loads the statistics (Refresh the newest and oldest log and count the log entries).
        /// </summary>
        public void LoadStatistics()
        {
            IPersistor persistor = new SQLitePersistor(this.databasePath);
            persistor.LoadStoredProcedure(StoredProcedures.SELECT_Statistics);
            using (DataSet dataSet = persistor.ExecuteQuery())
            {

                if (dataSet.Tables.Count == 0 || dataSet.Tables[0].Rows.Count == 0)
                {
                    this.count = 0;
                    this.minDate = this.maxDate = new DateTime();
                }
                else
                {
                    DataRow dataRow = dataSet.Tables[0].Rows[0];
                    this.count = (uint)((long)dataRow["Count"]);

                    this.minDate = (dataRow["MaxDate"].Equals(DBNull.Value))
                        ? new DateTime()
                        : DateTime.Parse((string)dataRow["MaxDate"]);

                    this.maxDate = (dataRow["MinDate"].Equals(DBNull.Value))
                        ? new DateTime()
                        : DateTime.Parse((string)dataRow["MinDate"]);
                }
            }
        }

        /// <summary>
        /// Gets the list from data set.
        /// </summary>
        /// <param name="dataSet">The data set.</param>
        /// <returns></returns>
        private static List<IRow> GetListFromDataSet(DataSet dataSet)
        {
            List<IRow> list = new List<IRow>();
            if (dataSet.Tables.Count <= 0)
                return null;

            foreach (DataRow dr in dataSet.Tables[0].Rows)
            {
                list.Add(GetLogRowFromDataRow(dr));
            }
            return list;
        }

        /// <summary>
        /// Gets the log from the specified DataRow.
        /// </summary>
        /// <param name="dataRow">The data row.</param>
        /// <returns>An instance of IRow</returns>
        private static IRow GetLogRowFromDataRow(DataRow dataRow)
        {
            if (!(dataRow["Type"] is long))
                throw new System.FormatException("Impossible to retreive the stamp of the log");

            long stampId = (long)dataRow["Type"];
            if ((--stampId) < 1)
                throw new System.FormatException("The value of the log's stamp is not in the range [1 < stamp < 7]");

            Log.Stamp stamp = (Log.Stamp)stampId;
            DateTime insertDate = ((DateTime)dataRow["InsertDate"]).ToLocalTime();
            string baseLogKey = (string)dataRow["fk_BaseLog"];
            string key = (string)dataRow["ID"];

            IRow logRow = new Row(new Guid(key), new Guid(baseLogKey), insertDate, stamp);
            logRow.Deepness = (int)(long)dataRow["Deepness"];
            logRow.ExceptionMessage = (dataRow["ExceptionMessage"].Equals(System.DBNull.Value)) ? null : (string)dataRow["ExceptionMessage"];
            logRow.Message = (dataRow["Message"].Equals(System.DBNull.Value)) ? null : (string)dataRow["Message"];
            logRow.Source = (dataRow["Source"].Equals(System.DBNull.Value)) ? null : (string)dataRow["Source"];
            logRow.StackTrace = (dataRow["StackTrace"].Equals(System.DBNull.Value)) ? null : (string)dataRow["StackTrace"];
            logRow.Title = (string)dataRow["Title"];

            return logRow;
        }

        #endregion Methods
    }

    internal static class StoredProcedures
    {
        #region Fields

        public const string SELECT_Dependency = "SELECT_Dependency";
        public const string SELECT_DistinctTitle = "SELECT_DistinctTitle";
        public const string SELECT_LogOnID = "SELECT_LogOnID";
        public const string SELECT_LogSorted = "SELECT_LogSorted";
        public const string SELECT_LogStored = "SELECT_LogStored";
        public const string SELECT_LogsBetweenDate = "SELECT_LogsBetweenDate";
        public const string SELECT_LogsByType = "SELECT_LogsByType";
        public const string SELECT_LogsOnExceptionMessageContains = "SELECT_LogsOnExceptionMessageContains";
        public const string SELECT_LogsOnMessageContains = "SELECT_LogsOnMessageContains";
        public const string SELECT_LogsOnTitleContains = "SELECT_LogsOnTitleContains";
        public const string SELECT_Statistics = "SELECT_Statistics";

        #endregion Fields
    }
}