/*
	Phoenix Suite 2
	Author: Jean-Baptiste Wautier
	Copyright 2006-2009 (c)
*/
namespace Phoenix.Log.Writer
{
    using System;
    using System.Collections.Generic;
    using System.Data;
    using System.Data.SQLite;
    using System.IO;
    using System.Reflection;

    using Phoenix.Log;
    using Phoenix.Log.Properties;

    /// <summary>
    /// Represent assembly stored procedure to access the log database
    /// </summary>
    public class SQLitePersistor : IPersistor
    {
        #region Fields

        private const string ConnectionStringTemplate = "Data Source={0};Version=3;";
        private const string SELECT_SP = "SELECT * FROM Query WHERE name = '{0}';";

        #endregion Fields

        #region Constructors

        /// <summary>
        /// Initializes assembly new instance of the <see cref="SQLitePersistor"/> class.
        /// This constructor will search the database errorPath in the default App.config.
        /// </summary>
        public SQLitePersistor()
            : this(Settings.Default.DbPath)
        {
        }

        /// <summary>
        /// Initializes assembly new instance of the <see cref="SQLitePersistor"/> class.
        /// </summary>
        /// <param name="databasePath">The errorPath of the database.</param>
        public SQLitePersistor(string databasePath)
        {
            this.Arguments = new Dictionary<string, string>();
            this.ConnectionString = string.Format(ConnectionStringTemplate, databasePath);

            if (!File.Exists(databasePath))
            {
                string backup = this.Query;
                try
                {
                    using (Stream stream = GetEmbeddedFile(typeof(SQLitePersistor).Assembly.GetName().Name, "Resources.db.sql"))
                    using (StreamReader reader = new StreamReader(stream))
                    {
                        this.Query = reader.ReadToEnd();
                    }
                    this.ExecuteNonQuery();
                }
                catch (Exception)
                {
                    throw;
                }
                this.Query = backup;
            }
        }

        #endregion Constructors

        #region Properties

        /// <summary>
        /// Gets or sets the arguments of the stored procedure.
        /// </summary>
        /// <value>The arguments.</value>
        public Dictionary<string, string> Arguments
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the query.
        /// </summary>
        /// <value>The query.</value>
        public string Query
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets or sets the connection string.
        /// </summary>
        /// <value>The connection string.</value>
        private string ConnectionString
        {
            get;
            set;
        }

        #endregion Properties

        #region Methods

        /// <summary>
        /// Executes assembly SQL query that does not return rows.
        /// Execute an UPDATE, DELETE or INSERT.
        /// </summary>
        /// <returns>
        /// 	<c>true</c> if the operation succeed. Otherwise <c>false</c>
        /// </returns>
        public bool ExecuteNonQuery()
        {
            this.SetArguments();
            int result;
            using (SQLiteConnection connection = new SQLiteConnection(ConnectionString))
            using (SQLiteCommand command = new SQLiteCommand(connection))
            {
                command.CommandType = CommandType.Text;
                command.CommandText = this.Query;

                connection.Open();
                result = command.ExecuteNonQuery();
                connection.Close();
            }
            if (result <= 0) return false;
            else return true;
        }

        /// <summary>
        /// Executes assembly SELECT query against the database.
        /// </summary>
        /// <returns>DataSet with the result of the query</returns>
        public DataSet ExecuteQuery()
        {
            this.SetArguments();
            SQLiteConnection connection = new SQLiteConnection(ConnectionString);
            SQLiteCommand command = new SQLiteCommand(connection);
            command.CommandText = this.Query;
            command.CommandType = CommandType.Text;

            DataSet dataSet = new DataSet();
            SQLiteDataAdapter dataAdapter = new SQLiteDataAdapter(command);
            dataAdapter.Fill(dataSet);

            return dataSet;
        }

        /// <summary>
        /// Execute the command and return the first column of the first row of the resultset
        /// (if present), or null if no resultset was returned.
        /// </summary>
        /// <returns>
        /// The first column of the first row of the first resultset from the query.
        /// </returns>
        public object ExecuteScalar()
        {
            this.SetArguments();

            using (SQLiteConnection connection = new SQLiteConnection(this.ConnectionString))
            using (SQLiteCommand command = new SQLiteCommand(connection))
            {
                command.CommandType = CommandType.Text;
                command.CommandText = this.Query;

                connection.Open();
                object result = command.ExecuteScalar();
                connection.Close();
                return (int)((long)result);
            }
        }

        /// <summary>
        /// Loads the stored procedure in memory.
        /// </summary>
        /// <returns></returns>
        public bool LoadStoredProcedure(string storedProcedureName)
        {
            int result;
            using (DataSet dataSet = new DataSet())
            {

                string sqlQuery =
                     string.Format(SELECT_SP, storedProcedureName);

                using (SQLiteConnection connection = new SQLiteConnection(ConnectionString))
                using (SQLiteCommand command = new SQLiteCommand(connection))
                {
                    command.CommandText = sqlQuery;
                    command.CommandType = CommandType.Text;

                    using (SQLiteDataAdapter dataAdapter = new SQLiteDataAdapter(command))
                    {
                        result = dataAdapter.Fill(dataSet);
                    }
                }

                if (result <= 0)
                {
                    return false;
                }
                else
                {
                    this.Query = (string)dataSet.Tables[0].Rows[0]["Query"];
                    return true;
                }
            }
        }

        /// <summary>
        /// Extracts an embedded textWriter out of assembly given assembly.
        /// </summary>
        /// <param name="assemblyName">The namespace of you assembly.</param>
        /// <param name="fileName">The name of the textWriter to extract.</param>
        /// <returns>A stream containing the textWriter data.</returns>
        private static Stream GetEmbeddedFile(string assemblyName, string fileName)
        {
            Stream str;

            if (assemblyName == null || string.IsNullOrEmpty(fileName))
            {
                string arg = (assemblyName == null) ? "assemblyName" : "fileName";
                throw new ArgumentNullException(arg
                    , string.Format("Impossible to have the embedded item because of bad arguments."));
            }

            Assembly assembly = Assembly.Load(assemblyName);
            str = assembly.GetManifestResourceStream(assemblyName + "." + fileName);

            if (str == null)
            {
                throw new ArgumentNullException(
                    string.Format("Could not locate embedded resource '{0}' in assembly '{1}'", fileName, assemblyName)
                    , "str");
            }
            return str;
        }

        /// <summary>
        /// Sets the arguments.
        /// </summary>
        private void SetArguments()
        {
            if (string.IsNullOrEmpty(this.Query))
            {
                throw new
                    LogException("This instance is not loaded. Please load a stored procedure before executing the query!");
            }

            foreach (KeyValuePair<string, string> item in this.Arguments)
            {
                this.Query = this.Query.Replace("@" + item.Key, item.Value);
            }
        }

        #endregion Methods
    }
}